/*
 * Copyright 2004-2006 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.myfaces.renderkit.html.util;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.HashMap;
import java.util.Map;

import javax.faces.context.FacesContext;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.myfaces.config.MyfacesConfig;
import org.apache.myfaces.util.ClassUtils;

/**
 * This class provides the ability to instantiate AddResource objects.
 * By default, this class will instantiate instances of
 * org.apache.myfaces.component.html.util.DefaultAddResource.
 * However, the context parameter org.apache.myfaces.ADD_RESOURCE_CLASS
 * can specify an alternative implementation of the AddResource
 * interface. The class must have a constructor with a single String
 * argument, representing the context path.
 * 
 * Mostly used to avoid having to include [script src="..."][/script]
 * in the head of the pages before using a component.
 * 
 * @author Peter Mahoney
 * @author Sylvain Vieujot (latest modification by $Author: mmarinschek $)
 * @version $Revision: 358042 $ $Date: 2005-12-20 17:12:56 +0000 (Tue, 20 Dec 2005) $
 */
public class AddResourceFactory
{

    protected static final Log log = LogFactory.getLog(AddResourceFactory.class);

    /**
     * Map of AddResource instances keyed by context path. This map will have
     * more than one entry only if the myfaces library is in the "shared"
     * classpath of a container where more than one webapp can see it
     * simultaneously.
     */
    private static final Map _addResourceMap = new HashMap();

    /**
     * Internal factory method.
     * <p>
     * Return an instance of AddResource keyed by context path, or create one
     * if no such instance already exists. Note that the same AddResource object
     * is shared among all threads servicing the same webapp, so all methods
     * on this class are required to be threadsafe.
     * <p>
     * Note that this method is package-scope for the purposes of unit-testing only.
     * This method should be treated as private by non-test code.
     */
    static AddResource getInstance(String contextPath, String addResourceClassName)
    {
        // Yes, this method does use a variant of the "double locking" idiom
        // which is well documented to be invalid for general use. However
        // it is believed safe for use here because AddResource objects in
        // _addResourceMap are never removed or replaced.
        AddResource instance = (AddResource) _addResourceMap.get(contextPath);
        if (instance == null)
        {
            synchronized (_addResourceMap)
            {
                instance = (AddResource) _addResourceMap.get(contextPath);
                if (instance == null)
                {
                	if (addResourceClassName == null)
                	{
                		// For efficiency don't use reflection unless it is necessary
                        instance = new DefaultAddResource(contextPath);                		
                	}
                	else
                	{
                		try
                		{
                			Class addResourceClass = ClassUtils.classForName(addResourceClassName);
                			
                			if (AddResource.class.isAssignableFrom(addResourceClass)) 
                			{
                				// Look for an appropriate constructor
                				Constructor[] constructors = addResourceClass.getDeclaredConstructors();
                				
                				for (int i = 0; i < constructors.length; i++)
                				{
                					Class[] parameters = constructors[i].getParameterTypes();

                					if (parameters.length == 1 && String.class.isAssignableFrom(parameters[i])) 
                					{
                						instance = (AddResource) constructors[i].newInstance(new String[] {contextPath});
                						log.info("Found AddResource class '" + addResourceClass.getName() + "'");
                						break;
                					}
                				}
                				
                				if (instance == null)
                				{
                        			log.error("Invalid AddResource class (" + addResourceClass.getName() 
                    					+ "). Must include single argument (context path) constructor.");                					
                				}
                			}
                			else
                			{
                    			log.error("Invalid AddResource class (" + addResourceClass.getName() 
                    					+ "). Must implement the AddResource interface.");
                			}
                		}
                		catch (ClassNotFoundException e)
                		{
                			log.error("AddResource class not found. Using default class instead", e);
                		} 
                		catch (IllegalArgumentException e) {
                			// This should not happen as the constructor has been checked
                			log.error(e);
						} 
                		catch (InstantiationException e) {
                			log.error("Invalid AddResource class. Must be non-abstract", e);
						} 
                		catch (IllegalAccessException e) {
                			log.error("Could not access AddResource class", e);
						} 
                		catch (InvocationTargetException e) {
                			log.error(e);
						}
						finally
						{
							// Ensure there is always an AddResource object available
							if (instance == null)
							{
								instance = new DefaultAddResource(contextPath);                			                				
							}
						}
                	}
                    _addResourceMap.put(contextPath, instance);
                }
            }
        }
        return instance;
    }

    public static AddResource getInstance(FacesContext context)
    {	
        return getInstance(context.getExternalContext().getRequestContextPath(),
        		MyfacesConfig.getCurrentInstance(context.getExternalContext()).getAddResourceClass());
    }

}
