/*
 * Copyright 2004-2006 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.myfaces.renderkit.html.util;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.myfaces.config.MyfacesConfig;
import org.apache.myfaces.renderkit.RendererUtils;
import org.apache.myfaces.renderkit.html.HTML;
import org.apache.myfaces.renderkit.html.HtmlRendererUtils;

import javax.faces.application.StateManager;
import javax.faces.application.ViewHandler;
import javax.faces.context.FacesContext;
import javax.faces.context.ResponseWriter;
import java.io.IOException;
import java.util.HashSet;
import java.util.Set;

/**
 * Many JSF components can be used without an enclosing h:form. In this
 * case, they need a dummy form to store data into and submit when
 * communication with the server is necessary. These components can use
 * methods on the <code>DummyFormRequestInfo</code> object accessable via this
 * class to register parameters and get the name of a form to submit.
 * <p>
 * Only one dummy form will be rendered into the response.
 * 
 * @author Manfred Geiler (latest modification by $Author: mmarinschek $)
 * @author Bruno Aranda
 * @version $Revision: 373532 $ $Date: 2006-01-30 15:47:34 +0000 (Mon, 30 Jan 2006) $
 */
public class DummyFormUtils
{
    private static final Log log = LogFactory.getLog(DummyFormUtils.class);

    public static final String DUMMY_FORM_NAME = "linkDummyForm";
    private static final String DUMMY_FORM_ID = "linkDummyForm";

    /**
     * Used to store the instance of <code>DummyFormRequestInfo</code> in the request map
     */
    public static final String DUMMY_FORM_INFO = DummyFormUtils.class.getName()+".DUMMY_FORM_INFO";


    public static String getDummyFormName() {
        return DUMMY_FORM_NAME;
    }

    /**
     * When writeDummyForm is set to true, a <code>DummyFormRequestInfo</code> object will be instantiated
     * and put in the request. Later, if this object is found, the dummyForm javascript will be rendered in the page
     * before the end of the <code>body</code> tag
     *
     * @param facesContext
     * @param writeDummyForm
     */
    public static void setWriteDummyForm(FacesContext facesContext, boolean writeDummyForm)
    {
        if (!writeDummyForm)
        {
            return;
        }

         if (!isWriteDummyForm(facesContext))
         {
             DummyFormRequestInfo dummyFormInfo = new DummyFormRequestInfo();
             facesContext.getExternalContext().getRequestMap().put(DUMMY_FORM_INFO, dummyFormInfo);
         }
    }

    /**
     * Checks if the DummyFormRequestInfo is already in the request map.
     * @param facesContext
     * @return boolean true, if dummy form is to be written
     */
    public static boolean isWriteDummyForm(FacesContext facesContext)
    {
        return facesContext.getExternalContext().getRequestMap().containsKey(DUMMY_FORM_INFO);
    }

    /**
     * Delegator method to add a parameter to the DummyFormRequestInfo object in the request
     * 
     * @param facesContext
     * @param paramName
     */
    public static void addDummyFormParameter(FacesContext facesContext, String paramName)
    {
        if (isWriteDummyForm(facesContext))
        {
            DummyFormRequestInfo dummyFormInfo = (DummyFormRequestInfo) facesContext.getExternalContext().getRequestMap().get(DUMMY_FORM_INFO);
            dummyFormInfo.addDummyFormParameter(paramName);
        }
        else
        {
            if (log.isWarnEnabled())
            {
                log.warn("Dummy Form parameter was not added because dummy form is not written");
            }
        }
    }

    public static Set getDummyFormParameters(FacesContext facesContext) {
        if (isWriteDummyForm(facesContext))
        {
            DummyFormRequestInfo dummyFormInfo = (DummyFormRequestInfo) facesContext.getExternalContext().getRequestMap().get(DUMMY_FORM_INFO);
            return dummyFormInfo.getDummyFormParams();
        }

        return new HashSet();
    }


    public static void writeDummyForm(ResponseWriter writer,
                                      Set dummyFormParams) throws IOException
    {
        FacesContext facesContext = FacesContext.getCurrentInstance();
        ViewHandler viewHandler = facesContext.getApplication().getViewHandler();
        String viewId = facesContext.getViewRoot().getViewId();
        String actionURL = viewHandler.getActionURL(facesContext, viewId);

        //write out dummy form
        writer.startElement(HTML.FORM_ELEM, null);
        writer.writeAttribute(HTML.ID_ATTR, DUMMY_FORM_ID, null);
        writer.writeAttribute(HTML.NAME_ATTR, DUMMY_FORM_NAME, null);
        writer.writeAttribute(HTML.STYLE_ATTR, "display:inline", null);
        writer.writeAttribute(HTML.METHOD_ATTR, "post", null);
        writer.writeURIAttribute(HTML.ACTION_ATTR,
                                 facesContext.getExternalContext().encodeActionURL(actionURL),
                                 null);
        writer.flush();

        StateManager stateManager = facesContext.getApplication().getStateManager();
        HtmlRendererUtils.writePrettyLineSeparator(facesContext);
        if (stateManager.isSavingStateInClient(facesContext))
        {
            //render state parameters
            //TODO: Optimize saveSerializedView call, because serialized view is built twice!
            StateManager.SerializedView serializedView = stateManager.saveSerializedView(facesContext);
            stateManager.writeState(facesContext, serializedView);
        }
        else
        {
            writer.startElement(HTML.INPUT_ELEM, null);
            writer.writeAttribute(HTML.TYPE_ATTR, HTML.INPUT_TYPE_HIDDEN, null);
            writer.writeAttribute(HTML.NAME_ATTR, RendererUtils.SEQUENCE_PARAM, null);
            writer.writeAttribute(HTML.VALUE_ATTR, RendererUtils.getViewSequence(facesContext), null);
            writer.endElement(HTML.INPUT_ELEM);
        }

        if (MyfacesConfig.getCurrentInstance(facesContext.getExternalContext()).isAutoScroll())
        {
            JavascriptUtils.renderAutoScrollHiddenInput(facesContext,writer);
        }

        HtmlRendererUtils.writePrettyLineSeparator(facesContext);
        if (dummyFormParams != null)
        {
            HtmlRendererUtils.renderHiddenCommandFormParams(writer, dummyFormParams);
            HtmlRendererUtils.renderClearHiddenCommandFormParamsFunction(writer,
                                                                         DUMMY_FORM_NAME,
                                                                         dummyFormParams, null);
        }
        HtmlRendererUtils.writePrettyLineSeparator(facesContext);

        writer.endElement(HTML.FORM_ELEM);
    }


}
